//----------------------------------------------------------------------------------
// File:   MultiTextureTerrain.cpp
// Author: Bryan Dudash
// Email:  sdkfeedback@nvidia.com
// 
// Copyright (c) 2007 NVIDIA Corporation. All rights reserved.
//
// TO  THE MAXIMUM  EXTENT PERMITTED  BY APPLICABLE  LAW, THIS SOFTWARE  IS PROVIDED
// *AS IS*  AND NVIDIA AND  ITS SUPPLIERS DISCLAIM  ALL WARRANTIES,  EITHER  EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED  TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL  NVIDIA OR ITS SUPPLIERS
// BE  LIABLE  FOR  ANY  SPECIAL,  INCIDENTAL,  INDIRECT,  OR  CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION,  DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE  USE OF OR INABILITY  TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
//
//----------------------------------------------------------------------------------

#include "DXUT.h"
#include "MultiTextureTerrain.h"
#include "sdkmesh_old.h"
#include "SDKmisc.h"

MultiTextureTerrain::MultiTextureTerrain()
{
    m_pEffect = NULL;
    m_pInputLayout = NULL;
    m_pVertexBuffer = NULL;
    m_pIndexBuffer = NULL;
    m_pTerrainTexture = NULL;
    m_pTerrainTextureSRV = NULL;

    m_pTerrainChecker= NULL;
    m_pTerrainCheckerSRV = NULL;

    m_pTerrainNormal = NULL;
    m_pTerrainNormalSRV=  NULL;

    m_pTechnique = NULL;
    m_pmWorldViewProjection = NULL;
    m_pTextureArrayEffectVar = NULL;

    m_bUseChecker = false;
    m_bUseNormalMaps = true;
    m_bRenderUVs = false;

    m_iNumIndices = 0;
    m_iNumVertices = 0;
    m_fRadius = 0.f;
}

MultiTextureTerrain::~MultiTextureTerrain()
{
}

/*
    We load our terrain mesh here.  Also note, i generate multiple UV sets procedurally.  In a regular game scenario you'd have these authored...
*/
void MultiTextureTerrain::initialize(ID3D10Device *pd3dDevice)
{
    HRESULT hr;

    const D3D10_INPUT_ELEMENT_DESC loadingLayout[] =
    {
        { "POSITION",    0, DXGI_FORMAT_R32G32B32_FLOAT,        0, 0,  D3D10_INPUT_PER_VERTEX_DATA, 0 },
        { "NORMAL",        0, DXGI_FORMAT_R32G32B32_FLOAT,        0, 12, D3D10_INPUT_PER_VERTEX_DATA, 0 },
        { "TEXCOORD",    0, DXGI_FORMAT_R32G32_FLOAT,        0, 24, D3D10_INPUT_PER_VERTEX_DATA, 0 },
        { "TANGENT",    0, DXGI_FORMAT_R32G32B32_FLOAT,        0, 32, D3D10_INPUT_PER_VERTEX_DATA, 0 },
        { "BITANGENT",    0, DXGI_FORMAT_R32G32B32_FLOAT,        0, 44, D3D10_INPUT_PER_VERTEX_DATA, 0 },
    };

    struct LoadingVertex
    {
        D3DXVECTOR3 pos;
        D3DXVECTOR3 norm;
        D3DXVECTOR2 tex;
        D3DXVECTOR3 tangent;
        D3DXVECTOR3 binorm;
    };

    const D3D10_INPUT_ELEMENT_DESC layout[] =
    {
        { "POSITION",    0, DXGI_FORMAT_R32G32B32_FLOAT,        0, 0,  D3D10_INPUT_PER_VERTEX_DATA, 0 },
        { "NORMAL",        0, DXGI_FORMAT_R32G32B32_FLOAT,        0, 12, D3D10_INPUT_PER_VERTEX_DATA, 0 },
        { "TANGENT",    0, DXGI_FORMAT_R32G32B32_FLOAT,        0, 24, D3D10_INPUT_PER_VERTEX_DATA, 0 },
        { "BINORMAL",    0, DXGI_FORMAT_R32G32B32_FLOAT,        0, 36, D3D10_INPUT_PER_VERTEX_DATA, 0 },
        { "TEXCOORD",    0, DXGI_FORMAT_R32G32B32A32_FLOAT,    0, 48, D3D10_INPUT_PER_VERTEX_DATA, 0 },
        { "TEXCOORD",    1, DXGI_FORMAT_R32G32B32A32_FLOAT,    0, 64, D3D10_INPUT_PER_VERTEX_DATA, 0 },
    };

    // Use DXUT utility mesh to load some terrain geometry
    CDXUTMesh10    Mesh;
    Mesh.Create(pd3dDevice,L"StHelen.x",(D3D10_INPUT_ELEMENT_DESC*)loadingLayout,sizeof(loadingLayout)/sizeof(D3D10_INPUT_ELEMENT_DESC));

    m_iNumVertices = Mesh.m_pMesh10->GetVertexCount();
    m_iNumIndices = Mesh.m_pMesh10->GetFaceCount()*3;

    // Now make a new VB for our "fatter" vertices
    Vertex *pVertices = new Vertex[m_iNumVertices];
    LoadingVertex *pSourceVertices = NULL;
    SIZE_T loadingSize = 0;
    ID3DX10MeshBuffer *pSourceVertexBuffer = NULL;

    Mesh.m_pMesh10->GetVertexBuffer(0,&pSourceVertexBuffer);
    pSourceVertexBuffer->Map((void**)&pSourceVertices,&loadingSize);
    if(loadingSize != m_iNumVertices * sizeof(LoadingVertex)) V(E_FAIL);

    // Iterate over vertices adding extra texcoord information for multiple textures.
    //  $ NOTE in a normal game scenario, this would be authored and exported.
    for(int iVertex = 0;iVertex<(int)m_iNumVertices;iVertex++)
    {
        pVertices[iVertex].pos = pSourceVertices[iVertex].pos;
        pVertices[iVertex].norm = pSourceVertices[iVertex].norm;
        pVertices[iVertex].tangent = pSourceVertices[iVertex].tangent;
        pVertices[iVertex].binorm = pSourceVertices[iVertex].binorm;

        // maintain running max vertical offset form origin as radius
        if(fabs(pVertices[iVertex].pos.y) > m_fRadius) m_fRadius = fabs(pVertices[iVertex].pos.y);

        const float fSnowLine = 0.8f;
        const float fTerrainLine = 0.3f;
        const float fRocksLine = -0.1f;

        const float fSnowTerrainBandSize = 3.0f;
        const float fTerrainRocksBandSize = 0.5f;
        const float fRocksLavaBandSize = 1.f;

        float fTex0Alpha = 0.f;
        float fTex1Alpha = 0.f;
        float fTex0Index = 0.f;
        float fTex1Index = 1.f;

        if(pVertices[iVertex].pos.y >= fSnowLine)            // snow and terrain line
        {
            fTex0Index = 0.f;
            fTex1Index = 1.f;
            fTex0Alpha = min(1.f,(pVertices[iVertex].pos.y - fSnowLine)/fSnowTerrainBandSize);
            fTex1Alpha = 1.f - fTex0Alpha;
        }
        else if(pVertices[iVertex].pos.y >= fTerrainLine)    // rocks and terrain 
        {
            fTex0Index = 2.f;
            fTex1Index = 1.f;
            fTex1Alpha = min(1.f,(pVertices[iVertex].pos.y - fTerrainLine)/fTerrainRocksBandSize);
            fTex0Alpha = 1.f - fTex1Alpha;
        }
        else if(pVertices[iVertex].pos.y >= fRocksLine)        // lava and rocks
        {
            fTex0Index = 2.f;
            fTex1Index = 3.f;
            fTex0Alpha = min(1.f,(pVertices[iVertex].pos.y - fRocksLine)/fRocksLavaBandSize);
            fTex1Alpha = 1.f - fTex0Alpha;
        }
        else
        {
            fTex0Index = 2.f;
            fTex1Index = 3.f;
            fTex0Alpha = 0.f;
            fTex1Alpha = 1.f;
        }

        D3DXVECTOR2 uv = pSourceVertices[iVertex].tex;
        pVertices[iVertex].tex0 = D3DXVECTOR4(1.f*uv.x,1.f*uv.y,fTex0Index,fTex0Alpha);
        pVertices[iVertex].tex1 = D3DXVECTOR4(1.f*uv.x,1.f*uv.y,fTex1Index,fTex1Alpha);
    }
    pSourceVertexBuffer->Unmap();
    SAFE_RELEASE(pSourceVertexBuffer);

    D3D10_BUFFER_DESC bufferDesc;
    ZeroMemory(&bufferDesc,sizeof(D3D10_BUFFER_DESC));
    bufferDesc.ByteWidth = m_iNumVertices*sizeof(Vertex);
    bufferDesc.Usage = D3D10_USAGE_DEFAULT;
    bufferDesc.BindFlags = D3D10_BIND_VERTEX_BUFFER;

    D3D10_SUBRESOURCE_DATA subresourceData;
    subresourceData.pSysMem = (void*)pVertices;
    V( pd3dDevice->CreateBuffer( &bufferDesc, &subresourceData, &m_pVertexBuffer ) );

    delete [] pVertices;

    // Just copy out the IB, since it should be identical
    void *pData = NULL;
    ID3DX10MeshBuffer *pSourceIndexBuffer = NULL;
    Mesh.m_pMesh10->GetIndexBuffer(&pSourceIndexBuffer);
    pSourceIndexBuffer->Map(&pData,&loadingSize);
    if(loadingSize != m_iNumIndices*sizeof(DWORD)) V(E_FAIL);

    ZeroMemory(&bufferDesc,sizeof(D3D10_BUFFER_DESC));
    bufferDesc.ByteWidth = m_iNumIndices*sizeof(DWORD);
    bufferDesc.Usage = D3D10_USAGE_DEFAULT;
    bufferDesc.BindFlags = D3D10_BIND_INDEX_BUFFER;

    subresourceData.pSysMem = (void*)pData;
    V( pd3dDevice->CreateBuffer( &bufferDesc, &subresourceData, &m_pIndexBuffer ) );

    pSourceIndexBuffer->Unmap();
    SAFE_RELEASE(pSourceIndexBuffer);

    Mesh.Destroy();

    // Load our FX and technique, and create layout variable
    WCHAR str[MAX_PATH];
    V( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"MultiTextureTerrain.fx" ) );
    V( D3DX10CreateEffectFromFile( str, NULL, NULL, "fx_4_0", D3D10_SHADER_ENABLE_STRICTNESS, 0, pd3dDevice, NULL, NULL, &m_pEffect, NULL, &hr ) );

    m_pTechnique = m_pEffect->GetTechniqueByName("RenderTerrain");
    m_pZFirstTechnique = m_pEffect->GetTechniqueByName("RenderTerrainZOnly");

    // Our vars
    m_pTextureArrayEffectVar = m_pEffect->GetVariableBySemantic("TextureArray")->AsShaderResource();
    m_pNormalArrayEffectVar = m_pEffect->GetVariableBySemantic("TextureNormalArray")->AsShaderResource();

    m_pmWorldViewProjection = m_pEffect->GetVariableBySemantic("WorldViewProjection")->AsMatrix();
    m_pmWorld = m_pEffect->GetVariableBySemantic("World")->AsMatrix();
    m_pmWorldInverseTranspose = m_pEffect->GetVariableBySemantic("WorldInverseTranspose")->AsMatrix();

    m_pRenderUVsVar = m_pEffect->GetVariableByName("g_bRenderUVs")->AsScalar();
    m_pUseNormalMapsVar = m_pEffect->GetVariableByName("g_bRenderWithNormalMaps")->AsScalar();

    D3D10_PASS_DESC PassDesc;
    V( m_pTechnique->GetPassByIndex( 0 )->GetDesc( &PassDesc ) );
    V( pd3dDevice->CreateInputLayout( layout, sizeof(layout)/sizeof(D3D10_INPUT_ELEMENT_DESC), PassDesc.pIAInputSignature, PassDesc.IAInputSignatureSize, &m_pInputLayout ) );

    // Make our terrain texture array with some interesting maps
    LPCTSTR pTextureFiles[] = {L"..\\Media\\snow.dds",L"..\\Media\\terrain.dds",L"rocks.dds",L"lava.dds"};
    loadTextureArray(pd3dDevice,pTextureFiles,4,&m_pTerrainTexture,&m_pTerrainTextureSRV);

    // Load some checker textures for visualization
    LPCTSTR pCheckerTextureFiles[] = {L"..\\Media\\mosaic-white.dds",L"..\\Media\\mosaic-green.dds",L"mosaic-blue.dds",L"mosaic-red.dds"};
    loadTextureArray(pd3dDevice,pCheckerTextureFiles,4,&m_pTerrainChecker,&m_pTerrainCheckerSRV);

    // An array of normal maps to match the diffuse/albedo
    LPCTSTR pTextureNormalFiles[] = {L"..\\Media\\snow_normal.dds",L"..\\Media\\terrain_normal.dds",L"rocks_normal.dds", L"lava_normal.dds"};
    loadTextureArray(pd3dDevice,pTextureNormalFiles,3,&m_pTerrainNormal,&m_pTerrainNormalSRV);

}

//--------------------------------------------------------------------------------------
// Standard draw for a indexed triangle list. 
//--------------------------------------------------------------------------------------

void MultiTextureTerrain::draw(ID3D10Device *pd3dDevice,D3DXMATRIX &mView,D3DXMATRIX &mProj)
{
    UINT uiVertexStride = sizeof(Vertex);
    UINT uOffset = 0;
    pd3dDevice->IASetInputLayout(m_pInputLayout);
    pd3dDevice->IASetPrimitiveTopology(D3D10_PRIMITIVE_TOPOLOGY_TRIANGLELIST);
    pd3dDevice->IASetVertexBuffers(0,1,&m_pVertexBuffer,&uiVertexStride,&uOffset);
    pd3dDevice->IASetIndexBuffer(m_pIndexBuffer,DXGI_FORMAT_R32_UINT,0);

    // Set the effect variables for the rendering
    D3DXMATRIX mWorld;
    D3DXMATRIX mWorldIT;
    D3DXMATRIX mWorldViewProj;
    D3DXMatrixIdentity(&mWorld);
    D3DXMatrixInverse(&mWorldIT,0,&mWorld);
    D3DXMatrixTranspose(&mWorldIT,&mWorldIT);
    mWorldViewProj = mWorld * mView * mProj;
    m_pmWorldViewProjection->SetMatrix((float*)&mWorldViewProj);
    m_pmWorld->SetMatrix((float*)&mWorld);
    m_pmWorldInverseTranspose->SetMatrix((float*)&mWorldIT);

    // Texture Arrays for diffuse and normals
    m_pNormalArrayEffectVar->SetResource(m_pTerrainNormalSRV);

    if(!m_bUseChecker)
        m_pTextureArrayEffectVar->SetResource(m_pTerrainTextureSRV);
    else
        m_pTextureArrayEffectVar->SetResource(m_pTerrainCheckerSRV);

    m_pUseNormalMapsVar->SetBool(m_bUseNormalMaps);
    m_pRenderUVsVar->SetBool(m_bRenderUVs);

    // Render Z only first to engage Z cull on occluded surfaces.
    D3D10_TECHNIQUE_DESC techDesc;
    memset( &techDesc, 0, sizeof( D3D10_TECHNIQUE_DESC ) );
    m_pZFirstTechnique->GetDesc( &techDesc );
    for (UINT iPass = 0; iPass < techDesc.Passes; iPass++)
    {
        ID3D10EffectPass *pCurrentPass = m_pZFirstTechnique->GetPassByIndex( iPass );
        pCurrentPass->Apply( 0 );

        if(m_pIndexBuffer)
            pd3dDevice->DrawIndexed(m_iNumIndices,0,0);
        else
            pd3dDevice->Draw(m_iNumVertices,0);
            
    }

    // Now render with the texture array
    memset( &techDesc, 0, sizeof( D3D10_TECHNIQUE_DESC ) );
    m_pTechnique->GetDesc( &techDesc );
    for (UINT iPass = 0; iPass < techDesc.Passes; iPass++)
    {
        ID3D10EffectPass *pCurrentPass = m_pTechnique->GetPassByIndex( iPass );
        pCurrentPass->Apply( 0 );

        if(m_pIndexBuffer)
            pd3dDevice->DrawIndexed(m_iNumIndices,0,0);
        else
            pd3dDevice->Draw(m_iNumVertices,0);
    }
}

void MultiTextureTerrain::release()
{
    SAFE_RELEASE(m_pInputLayout);
    SAFE_RELEASE(m_pVertexBuffer);
    SAFE_RELEASE(m_pIndexBuffer);
    SAFE_RELEASE(m_pTerrainTexture);
    SAFE_RELEASE(m_pTerrainTextureSRV);
    SAFE_RELEASE(m_pTerrainChecker);
    SAFE_RELEASE(m_pTerrainCheckerSRV);
    SAFE_RELEASE(m_pTerrainNormal);
    SAFE_RELEASE(m_pTerrainNormalSRV);
    SAFE_RELEASE(m_pEffect);
}

float MultiTextureTerrain::getRadius()
{
    return m_fRadius;
}

//--------------------------------------------------------------------------------------
// LoadTextureArray loads a texture array and associated view from a series
// of textures on disk.  
//--------------------------------------------------------------------------------------
HRESULT MultiTextureTerrain::loadTextureArray( ID3D10Device* pd3dDevice, LPCTSTR* szTextureNames, int iNumTextures, ID3D10Texture2D** ppTex2D, ID3D10ShaderResourceView** ppSRV)
{
    HRESULT hr = S_OK;
    D3D10_TEXTURE2D_DESC desc;
    ZeroMemory( &desc, sizeof(D3D10_TEXTURE2D_DESC) );

    WCHAR str[MAX_PATH];
    for(int i=0; i<iNumTextures; i++)
    {
        V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, szTextureNames[i] ) );

        ID3D10Resource *pRes = NULL;
        D3DX10_IMAGE_LOAD_INFO loadInfo;
        ZeroMemory( &loadInfo, sizeof( D3DX10_IMAGE_LOAD_INFO ) );
        loadInfo.Width = D3DX_FROM_FILE;
        loadInfo.Height  = D3DX_FROM_FILE;
        loadInfo.Depth  = D3DX_FROM_FILE;
        loadInfo.FirstMipLevel = 0;
        loadInfo.MipLevels = D3DX_FROM_FILE;
        loadInfo.Usage = D3D10_USAGE_STAGING;
        loadInfo.BindFlags = 0;
        loadInfo.CpuAccessFlags = D3D10_CPU_ACCESS_WRITE | D3D10_CPU_ACCESS_READ;
        loadInfo.MiscFlags = 0;
        loadInfo.Format = DXGI_FORMAT_R8G8B8A8_UNORM;
        loadInfo.Filter = D3DX10_FILTER_NONE;
        loadInfo.MipFilter = D3DX10_FILTER_NONE;

        V_RETURN(D3DX10CreateTextureFromFile( pd3dDevice, str, &loadInfo, NULL, &pRes, &hr ));
        if( pRes )
        {
            ID3D10Texture2D* pTemp;
            pRes->QueryInterface( __uuidof( ID3D10Texture2D ), (LPVOID*)&pTemp );
            pTemp->GetDesc( &desc );

            D3D10_MAPPED_TEXTURE2D mappedTex2D;
            if(DXGI_FORMAT_R8G8B8A8_UNORM != desc.Format)   //make sure we're R8G8B8A8
                return false;

            if(desc.MipLevels > 4)
                desc.MipLevels -= 4;
            if(!(*ppTex2D))
            {
                desc.Usage = D3D10_USAGE_DEFAULT;
                desc.BindFlags = D3D10_BIND_SHADER_RESOURCE;
                desc.CPUAccessFlags = 0;
                desc.ArraySize = iNumTextures;
                V_RETURN(pd3dDevice->CreateTexture2D( &desc, NULL, ppTex2D));
            }

            for(UINT iMip=0; iMip < desc.MipLevels; iMip++)
            {
                pTemp->Map( iMip, D3D10_MAP_READ, 0, &mappedTex2D );

                pd3dDevice->UpdateSubresource( (*ppTex2D), 
                    D3D10CalcSubresource( iMip, i, desc.MipLevels ),
                    NULL,
                    mappedTex2D.pData,
                    mappedTex2D.RowPitch,
                    0 );

                pTemp->Unmap( iMip );
            }

            SAFE_RELEASE( pRes );
            SAFE_RELEASE( pTemp );
        }
        else
        {
            return false;
        }
    }

    D3D10_SHADER_RESOURCE_VIEW_DESC SRVDesc;
    ZeroMemory( &SRVDesc, sizeof(SRVDesc) );
    SRVDesc.Format = desc.Format;
    SRVDesc.ViewDimension = D3D10_SRV_DIMENSION_TEXTURE2DARRAY;
    SRVDesc.Texture2DArray.MipLevels = desc.MipLevels;
    SRVDesc.Texture2DArray.ArraySize = iNumTextures;
    V_RETURN(pd3dDevice->CreateShaderResourceView( *ppTex2D, &SRVDesc, ppSRV ));

    return hr;
}