//----------------------------------------------------------------------------------
// File:   TextureArrayTerrain.cpp
// Author: Bryan Dudash
// Email:  sdkfeedback@nvidia.com
// 
// Copyright (c) 2007 NVIDIA Corporation. All rights reserved.
//
// TO  THE MAXIMUM  EXTENT PERMITTED  BY APPLICABLE  LAW, THIS SOFTWARE  IS PROVIDED
// *AS IS*  AND NVIDIA AND  ITS SUPPLIERS DISCLAIM  ALL WARRANTIES,  EITHER  EXPRESS
// OR IMPLIED, INCLUDING, BUT NOT LIMITED  TO, IMPLIED WARRANTIES OF MERCHANTABILITY
// AND FITNESS FOR A PARTICULAR PURPOSE.  IN NO EVENT SHALL  NVIDIA OR ITS SUPPLIERS
// BE  LIABLE  FOR  ANY  SPECIAL,  INCIDENTAL,  INDIRECT,  OR  CONSEQUENTIAL DAMAGES
// WHATSOEVER (INCLUDING, WITHOUT LIMITATION,  DAMAGES FOR LOSS OF BUSINESS PROFITS,
// BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR ANY OTHER PECUNIARY LOSS)
// ARISING OUT OF THE  USE OF OR INABILITY  TO USE THIS SOFTWARE, EVEN IF NVIDIA HAS
// BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
//
//----------------------------------------------------------------------------------

// The wrapper app around the MultiTexture terrain class.

#include "DXUT.h"
#include "MultiTextureTerrain.h"
#include "DXUTcamera.h"
#include "SDKmisc.h"
#include "DXUTgui.h"
#include "DXUTSettingsDlg.h"

MultiTextureTerrain        g_Terrain;                // our multi texture terrain interface

CFirstPersonCamera      g_Camera;               
ID3DX10Font*            g_pFont10 = NULL;       
ID3DX10Sprite*          g_pSprite10 = NULL;
CDXUTTextHelper*        g_pTxtHelper = NULL;
CDXUTDialogResourceManager   g_DialogResourceManager; // manager for shared resources of dialogs
CDXUTDialog             g_SampleUI;             // dialog for sample specific controls
CDXUTDialog             g_HUD;
CD3DSettingsDlg         g_SettingsDlg;
bool                    g_bShowHelp;

#define IDC_TOGGLEFULLSCREEN        1
#define IDC_TOGGLEREF               2
#define IDC_CHANGEDEVICE            3

#define IDC_ENABLENORMALMAPS        4
#define IDC_SHOWUVS                 5
#define IDC_MOSAIC                  6

void    CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext );

//--------------------------------------------------------------------------------------
// Reject any D3D10 devices that aren't acceptable by returning false
//--------------------------------------------------------------------------------------
bool CALLBACK IsD3D10DeviceAcceptable( UINT Adapter, UINT Output, D3D10_DRIVER_TYPE DeviceType, DXGI_FORMAT BackBufferFormat, bool bWindowed, void* pUserContext )
{
    return true;
}


//--------------------------------------------------------------------------------------
// Called right before creating a D3D9 or D3D10 device, allowing the app to modify the device settings as needed
//--------------------------------------------------------------------------------------
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, void* pUserContext )
{
    return true;
}


//--------------------------------------------------------------------------------------
// Create any D3D10 resources that aren't dependant on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D10CreateDevice( ID3D10Device* pd3dDevice, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
    HRESULT hr;

    g_SettingsDlg.GetDialogControl()->RemoveAllControls();
    g_SettingsDlg.Init( &g_DialogResourceManager );

    V_RETURN( g_DialogResourceManager.OnD3D10CreateDevice( pd3dDevice ) );
    V_RETURN( g_SettingsDlg.OnD3D10CreateDevice( pd3dDevice ) );
    V_RETURN( D3DX10CreateFont( pd3dDevice, 15, 0, FW_BOLD, 1, FALSE, DEFAULT_CHARSET, 
        OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, 
        L"Arial", &g_pFont10 ) );
    V_RETURN( D3DX10CreateSprite( pd3dDevice, 512, &g_pSprite10 ) );
    g_pTxtHelper = new CDXUTTextHelper( NULL, NULL, g_pFont10, g_pSprite10, 15 );

    // Initialize our mesh and textures, creates DX10 resources, loads shaders, etc.
    g_Terrain.initialize(pd3dDevice);

    // Setup the camera's view parameters
    D3DXVECTOR3 vecEye(11.0f, 6.0f, -0.f);
    D3DXVECTOR3 vecAt (0.0f, 0.0f, 0.0f);
    g_Camera.SetViewParams( &vecEye, &vecAt );
    g_Camera.SetEnablePositionMovement(true);
    g_Camera.SetScalers(0.01f,1.f);

//    FLOAT fCameraRadius = 2.f*g_Terrain.getRadius();
//    g_Camera.SetRadius( fCameraRadius, fCameraRadius, fCameraRadius );

    return S_OK;
}


//--------------------------------------------------------------------------------------
// Create any D3D10 resources that depend on the back buffer
//--------------------------------------------------------------------------------------
HRESULT CALLBACK OnD3D10ResizedSwapChain( ID3D10Device* pd3dDevice, IDXGISwapChain *pSwapChain, const DXGI_SURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext )
{
    HRESULT hr;

    // Setup the camera's projection parameters
    float fAspectRatio = pBackBufferSurfaceDesc->Width / (FLOAT)pBackBufferSurfaceDesc->Height;
    g_Camera.SetProjParams( D3DX_PI/4, fAspectRatio, 0.1f, 400.0f );

    // Box is at the bottom right
    const int width = 150;
    const int height = 70;
    g_SampleUI.SetLocation( pBackBufferSurfaceDesc->Width-width, pBackBufferSurfaceDesc->Height-height );
    g_SampleUI.SetSize( width, height );

    g_HUD.SetLocation( pBackBufferSurfaceDesc->Width - 170, 0);
    g_HUD.SetSize( 170, 170 );

    V_RETURN( g_DialogResourceManager.OnD3D10ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc ) );
    V_RETURN( g_SettingsDlg.OnD3D10ResizedSwapChain( pd3dDevice, pBackBufferSurfaceDesc ) );

    return S_OK;
}


//--------------------------------------------------------------------------------------
// Handle updates to the scene.  This is called regardless of which D3D API is used
//--------------------------------------------------------------------------------------
void CALLBACK OnFrameMove( double fTime, float fElapsedTime, void* pUserContext )
{
    g_Camera.FrameMove( fElapsedTime );
}


//--------------------------------------------------------------------------------------
// Render the help and statistics text
//--------------------------------------------------------------------------------------
void RenderText()
{
    g_pTxtHelper->Begin();
    g_pTxtHelper->SetInsertionPos( 2, 0 );
    g_pTxtHelper->SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 0.0f, 1.0f ) );
    g_pTxtHelper->DrawTextLine( DXUTGetFrameStats(true) );
    g_pTxtHelper->DrawTextLine( DXUTGetDeviceStats() );
    if(g_bShowHelp)
    {
        g_pTxtHelper->DrawTextLine( L"Arrow keys move camera.  ");
        g_pTxtHelper->DrawTextLine( L"Mouse Left Click and Drag rotates camera." );
        g_pTxtHelper->DrawTextLine( L"UV Visualization:");
        g_pTxtHelper->DrawTextLine( L"  Texcoord 0 = red ");
        g_pTxtHelper->DrawTextLine( L"  Texcoord 1 = green " );
        g_pTxtHelper->DrawTextLine( L"Mosaic conforms textures to same pattern, ");
        g_pTxtHelper->DrawTextLine( L"  changing only color between seperate layers." );
    }
    else
    {
        g_pTxtHelper->DrawTextLine( L"Hit F1 for Help." );
    }
    g_pTxtHelper->End();
}

//--------------------------------------------------------------------------------------
// Render the scene using the D3D10 device
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10FrameRender( ID3D10Device* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext )
{
    // Clear render target and the depth stencil
    //float ClearColor[4] = { 0.176f, 0.196f, 0.667f, 0.0f };
    float ClearColor[4] = { 0.f, 0.f, 0.f, 0.0f };
    pd3dDevice->ClearRenderTargetView( DXUTGetD3D10RenderTargetView(), ClearColor );
    pd3dDevice->ClearDepthStencilView( DXUTGetD3D10DepthStencilView(), D3D10_CLEAR_DEPTH, 1.0, 0 );

    if( g_SettingsDlg.IsActive() )
    {
        g_SettingsDlg.OnRender( fElapsedTime );
        return;
    }

    D3DXMATRIX  mView = *g_Camera.GetViewMatrix();
    D3DXMATRIX  mProj = *g_Camera.GetProjMatrix();

    g_Terrain.draw(pd3dDevice,mView,mProj);

    // Render GUI

    g_HUD.OnRender( fElapsedTime );

    RenderText();
    g_SampleUI.OnRender( fElapsedTime ); 

    /*ID3D10Resource *pDst = NULL;
    DXUTGetD3D10RenderTargetView()->GetResource(&pDst);
    WCHAR filename[100];
    StringCchPrintf(filename, 100, L"screenshot.dds"); 
    D3DX10SaveTextureToFile(pDst, D3DX10_IFF_DDS, filename);
    pDst->Release();*/

}


//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10ResizedSwapChain
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10ReleasingSwapChain( void* pUserContext )
{
    g_DialogResourceManager.OnD3D10ReleasingSwapChain();
}


//--------------------------------------------------------------------------------------
// Release D3D10 resources created in OnD3D10CreateDevice
//--------------------------------------------------------------------------------------
void CALLBACK OnD3D10DestroyDevice( void* pUserContext )
{
    g_DialogResourceManager.OnD3D10DestroyDevice();
    g_SettingsDlg.GetDialogControl()->RemoveAllControls();
    g_SettingsDlg.OnD3D10DestroyDevice();
    g_SampleUI.RemoveAllControls();
    g_HUD.RemoveAllControls();
    g_Terrain.release();

    SAFE_RELEASE( g_pFont10 );
    SAFE_RELEASE( g_pSprite10 );
    SAFE_DELETE( g_pTxtHelper );
}


//--------------------------------------------------------------------------------------
// Handle messages to the application
//--------------------------------------------------------------------------------------
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam,
                          bool* pbNoFurtherProcessing, void* pUserContext )
{
    if( g_SettingsDlg.IsActive() )
    {
        g_SettingsDlg.MsgProc( hWnd, uMsg, wParam, lParam );
        return 0;
    }

    if(g_SampleUI.MsgProc(hWnd,uMsg,wParam,lParam))
    {
        return 1;
    }

    if( g_HUD.MsgProc( hWnd, uMsg, wParam, lParam ) )
    {
        return 1;
    }

    if(g_Camera.HandleMessages( hWnd, uMsg, wParam, lParam ))
    {
        return 1;
    }

    if(uMsg == WM_KEYDOWN)
    {
        if((char)wParam == VK_F1)
        {
            g_bShowHelp = !g_bShowHelp;
        }
        return 1;
    }

    return 0;
}


//--------------------------------------------------------------------------------------
// Handle key presses
//--------------------------------------------------------------------------------------
void CALLBACK OnKeyboard( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext )
{
}


//--------------------------------------------------------------------------------------
// Handle mouse button presses
//--------------------------------------------------------------------------------------
void CALLBACK OnMouse( bool bLeftButtonDown, bool bRightButtonDown, bool bMiddleButtonDown,
                       bool bSideButton1Down, bool bSideButton2Down, int nMouseWheelDelta,
                       int xPos, int yPos, void* pUserContext )
{
}


//--------------------------------------------------------------------------------------
// Call if device was removed.  Return true to find a new device, false to quit
//--------------------------------------------------------------------------------------
bool CALLBACK OnDeviceRemoved( void* pUserContext )
{
    return true;
}

//--------------------------------------------------------------------------------------
// Handles the GUI events
//--------------------------------------------------------------------------------------
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext )
{
    switch( nControlID )
    {
    case IDC_TOGGLEFULLSCREEN: PostMessageW(DXUTGetHWND(), WM_SYSKEYDOWN, VK_RETURN, (1 << 29)); break;
    case IDC_TOGGLEREF:        DXUTToggleREF(); break;
    case IDC_CHANGEDEVICE:     g_SettingsDlg.SetActive( !g_SettingsDlg.IsActive() ); break;

    case IDC_ENABLENORMALMAPS:
        {
            g_Terrain.m_bUseNormalMaps = static_cast<CDXUTCheckBox*>(pControl)->GetChecked();
        }
        break;
    case IDC_SHOWUVS:
        {
            g_Terrain.m_bRenderUVs = static_cast<CDXUTCheckBox*>(pControl)->GetChecked();
            if(g_Terrain.m_bRenderUVs)
            {
                g_Terrain.m_bUseChecker = false;
                g_SampleUI.GetCheckBox(IDC_MOSAIC)->SetChecked(false);
            }
        }
        break;
    case IDC_MOSAIC:
        {
            g_Terrain.m_bUseChecker = static_cast<CDXUTCheckBox*>(pControl)->GetChecked();
            if(g_Terrain.m_bUseChecker)
            {
                g_Terrain.m_bRenderUVs = false;
                g_SampleUI.GetCheckBox(IDC_SHOWUVS)->SetChecked(false);
            }
        }
        break;

    default:
        break;
    }
}

void AppInit()
{
    int iX = 0;
    int iY = 0;
    g_SampleUI.Init( &g_DialogResourceManager );
    g_SampleUI.SetCallback( OnGUIEvent );

    g_SampleUI.AddCheckBox(IDC_MOSAIC,L"Mosaic Textures",iX,iY,150,12);
    g_SampleUI.AddCheckBox(IDC_SHOWUVS,L"UV Visualization",iX,iY+=14,150,12);
    g_SampleUI.AddCheckBox(IDC_ENABLENORMALMAPS,L"Enable Normal Mapping",iX,iY+=14,150,12,true);

    g_SettingsDlg.Init( &g_DialogResourceManager );
    g_HUD.Init( &g_DialogResourceManager );

    g_HUD.SetCallback( OnGUIEvent ); iY = 10;
    g_HUD.AddButton( IDC_TOGGLEFULLSCREEN, L"Toggle full screen", 35, iY, 125, 22 );
    g_HUD.AddButton( IDC_TOGGLEREF, L"Toggle REF (F3)", 35, iY += 24, 125, 22, VK_F3 );
    g_HUD.AddButton( IDC_CHANGEDEVICE, L"Change device (F2)", 35, iY += 24, 125, 22, VK_F2 );
    

}

//--------------------------------------------------------------------------------------
// Initialize everything and go into a render loop
//--------------------------------------------------------------------------------------
int WINAPI wWinMain( HINSTANCE hInstance, HINSTANCE hPrevInstance, LPWSTR lpCmdLine, int nCmdShow )
{
    // Enable run-time memory check for debug builds.
#if defined(DEBUG) | defined(_DEBUG)
    _CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

    // DXUT will create and use the best device (either D3D9 or D3D10)
    // that is available on the system depending on which D3D callbacks are set below

    // Set general DXUT callbacks
    DXUTSetCallbackFrameMove( OnFrameMove );
    DXUTSetCallbackKeyboard( OnKeyboard );
    DXUTSetCallbackMouse( OnMouse );
    DXUTSetCallbackMsgProc( MsgProc );
    DXUTSetCallbackDeviceChanging( ModifyDeviceSettings );
    DXUTSetCallbackDeviceRemoved( OnDeviceRemoved );

   // Set the D3D10 DXUT callbacks. Remove these sets if the app doesn't need to support D3D10
    DXUTSetCallbackD3D10DeviceAcceptable( IsD3D10DeviceAcceptable );
    DXUTSetCallbackD3D10DeviceCreated( OnD3D10CreateDevice );
    DXUTSetCallbackD3D10SwapChainResized( OnD3D10ResizedSwapChain );
    DXUTSetCallbackD3D10FrameRender( OnD3D10FrameRender );
    DXUTSetCallbackD3D10SwapChainReleasing( OnD3D10ReleasingSwapChain );
    DXUTSetCallbackD3D10DeviceDestroyed( OnD3D10DestroyDevice );

    DXUTSetHotkeyHandling(false,true,true);

    // Perform any application-level initialization here
    AppInit();

    HRESULT hr;
    V_RETURN(DXUTSetMediaSearchPath(L"..\\Source\\TextureArrayTerrain"));

    DXUTInit( true, true, NULL ); // Parse the command line, show msgboxes on error, no extra command line params
    DXUTSetCursorSettings( true, true ); // Show the cursor and clip it when in full screen
    DXUTCreateWindow( L"TextureArrayTerrain" );
    DXUTCreateDevice( true, 640, 480 );
    DXUTMainLoop(); // Enter into the DXUT render loop

    // Perform any application-level cleanup here

    return DXUTGetExitCode();
}


